from __future__ import annotations  # for python 3.8

from typing import Final

from agstoolbox.core.version.version import Version
from agstoolbox.core.version.version_utils import version_str_to_version, is_version_a_bigger_than_b


def get_installable_ags_versions() -> list[str]:
    versions: Final[list[str]] = [
        '3.4.3',
        '3.5.0',
        '3.5.1',
        '3.6.0',
        '3.6.1',
        '3.6.2',
        '3.99.99',
        '3.99.100',
        '4.0.0']

    return versions

def is_ags4(version: Version) -> bool:
    # there is an off-by-one mistake here, but let's leave this for now
    ags4_first: Version =  version_str_to_version('3.99.99')
    return is_version_a_bigger_than_b(version, ags4_first)



from __future__ import annotations  # for python 3.8

import fnmatch
import os
import re
from shutil import move as shutil_move
from typing import List
from pathlib import Path

from agstoolbox.core.ags.ags_editor_versions import is_ags4
from agstoolbox.core.ags.datafile_writer import get_multifile_lib, make_data_file_from_multifile_lib
from agstoolbox.core.ags.game_project import GameProject
from agstoolbox.core.ags.multifilelib import MultiFileLib
from agstoolbox.core.utils.file import join_paths_as_posix, get_file_if_exists, get_absolute_path, \
    get_posix
from agstoolbox.core.version.version import Version
from agstoolbox.core.version.version_utils import version_str_to_version, \
    is_version_a_bigger_than_b, is_version_a_between_bc


def case_insensitive_glob(directory: str, file_mask: str) -> list[Path]:
    reg_expr = re.compile(fnmatch.translate(file_mask), re.IGNORECASE)
    list_path = [i for i in os.listdir(directory) if os.path.isfile(os.path.join(directory, i))]
    result = [os.path.join(directory, j) for j in list_path if re.match(reg_expr, j)]
    result = [Path(j) for j in result]
    if result is None:
        return []
    return result


def get_dir_file_list(directory: str, file_mask: str):
    try:
        path = Path(directory).as_posix()
        return case_insensitive_glob(path, file_mask)
    except IOError:
        return []


def add_matching_files(file_list: List[str], file_mask: str, parent_dir: str,
                       full_paths: bool = False):
    files = get_dir_file_list(parent_dir, file_mask)
    if full_paths:
        file_list.extend([str(file) for file in files])
    else:
        file_list.extend([str(file.relative_to(parent_dir)) for file in files])


def construct_template_filelist(parent_dir: str) -> List[str]:
    files_to_include: List[str] = list()
    add_matching_files(files_to_include, "*.ico", parent_dir)
    add_matching_files(files_to_include, "Game.agf", parent_dir)
    add_matching_files(files_to_include, "acsprset.spr", parent_dir)
    add_matching_files(files_to_include, "preload.pcx", parent_dir)
    add_matching_files(files_to_include, "AudioCache/*.*", parent_dir)
    add_matching_files(files_to_include, "Speech/*.*", parent_dir)
    add_matching_files(files_to_include, "flic*.fl?", parent_dir)
    add_matching_files(files_to_include, "agsfnt*.ttf", parent_dir)
    add_matching_files(files_to_include, "agsfnt*.wfn", parent_dir)
    add_matching_files(files_to_include, "*.crm", parent_dir)
    add_matching_files(files_to_include, "*.asc", parent_dir)
    add_matching_files(files_to_include, "*.ash", parent_dir)
    add_matching_files(files_to_include, "*.txt", parent_dir)
    add_matching_files(files_to_include, "*.trs", parent_dir)
    add_matching_files(files_to_include, "*.pdf", parent_dir)
    add_matching_files(files_to_include, "*.ogv", parent_dir)
    return files_to_include


def create_template_from_game_dir(project_dir: str, template_filepath: str):
    if Path(template_filepath).exists():
        Path(template_filepath).unlink()

    template_filename: str = Path(template_filepath).name
    template_dir: str = str(Path(template_filepath).parent.as_posix())

    template_filelist: List[str] = construct_template_filelist(project_dir)
    filelist: List[str] = list()
    for file in template_filelist:
        filelist.append(join_paths_as_posix(project_dir, file))

    mlib: MultiFileLib = get_multifile_lib(filelist, project_dir, template_filename, False)
    make_data_file_from_multifile_lib(mlib, template_dir)


# the command below builds using a reimplemented template export
# the one using the editor is in ags_local_run
def create_template_from_project(game_project: GameProject, template_name: str, out: str | None):
    if out is None:
        out = game_project.directory

    template_name = Path(template_name).name
    template_filepath: str = join_paths_as_posix(out, template_name)
    create_template_from_game_dir(game_project.directory, template_filepath)


# Note, in ags
# v4.0.0.9, v4.0.0.10 ...
# and
# v3.6.2.1, v3.6.2.2 ...
# /maketemplate command is supported in the Editor
def editor_supports_template_export(game_project: GameProject) -> bool:
    ags3_first_sup: Version = version_str_to_version('3.6.2.1')
    ags4_first_sup: Version = version_str_to_version('4.0.0.9')
    ags3_to_ags4_ver: Version = version_str_to_version('3.99.0.0')
    ags4: bool = is_ags4(game_project.ags_editor_version)
    if ags4:
        return is_version_a_bigger_than_b(game_project.ags_editor_version, ags4_first_sup)
    else:
        return is_version_a_between_bc(game_project.ags_editor_version, ags3_first_sup, ags3_to_ags4_ver)


# Template built with editor uses instead GameFileName + ".agt"
def fix_editor_built_template(game_project: GameProject, out_dir: str, filename: str) -> bool:
    built_dir: str = game_project.directory
    built_file: str = game_project.game_file + ".agt"
    built_filename = get_file_if_exists(built_dir, built_file)
    if built_filename is None:
        return False

    built_filename = get_posix(get_absolute_path(built_filename))
    target_filename: str = get_posix(get_absolute_path(os.path.join(out_dir, filename)))
    if target_filename == built_filename:
        # no need to do anything
        return True

    existing_target_filename: str = get_file_if_exists(out_dir, filename)
    if existing_target_filename is not None:
        shutil_move(existing_target_filename, existing_target_filename + ".bkp")

    shutil_move(built_filename, target_filename)
    return True



from __future__ import annotations  # for python 3.8

import os
import sys
from subprocess import Popen, PIPE, TimeoutExpired

from agstoolbox.core.utils.file import get_file, get_dir
from agstoolbox.core.utils.pyinstaller_hacks import lock_dll_dir, unlock_dll_dir


def run_exe_params(exe_path: str, block: bool = False, timeout: int = 0, params=None) -> int:
    """
    Runs an executable file as a subprocess with optional blocking and timeout behavior.

    Args:
        exe_path (str): The path to the executable file to run.
        block (bool): If True, waits for the process to finish, respecting the timeout.
                      If False, the function returns immediately after starting the process.
        timeout (int): The maximum time to wait for the process to complete, in seconds.
                       If set to 0 or less, a default timeout of 10000 seconds is used.
                       Only valid when blocking.
        params (list, optional): Additional command-line parameters to pass to the executable.

    Returns:
        int: The return code of the process if it completes within the timeout.
             Returns -1 if the process is terminated due to exceeding the timeout.
    """
    if params is None:
        params = []

    exe_file = get_file(exe_path)
    working_dir = get_dir(exe_path)

    p_params = [exe_file]
    p_params.extend(params)

    # timeout is given in seconds
    if timeout <= 0:
        timeout = 10000

    cwd = os.getcwd()
    os.chdir(working_dir)
    lock_dll_dir()
    print('Popen: cwd=' + working_dir + ', ' + ' '.join(p_params))
    proc = Popen(p_params, cwd=working_dir, stdout=PIPE, stderr=PIPE, bufsize=1,
                 universal_newlines=True)

    timed_out: bool = False
    if block:
        try:
            # Wait for the process to complete or timeout
            outs, errs = proc.communicate(timeout=timeout)

            if outs:
                sys.stdout.write(outs)
            if errs:
                sys.stderr.write(errs)

        except TimeoutExpired:
            print("ERROR: Timeout time exceeded!")
            timed_out = True
        except Exception as e:
            print(f"ERROR: An exception occurred while running the process: {e}")
            timed_out = True

    return_code: int = proc.returncode
    if timed_out:
        proc.terminate()
        return_code = -1

    unlock_dll_dir()
    os.chdir(cwd)
    return return_code



from __future__ import annotations  # for python 3.8


class MultiFile:
    name: str | None = None
    filename: str | None = None
    datafile: int | None = None
    length: int | None = None
    offset: int | None = None



from __future__ import annotations  # for python 3.8

from agstoolbox.core.ags.multifile import MultiFile


class MultiFileLib:
    data_file_names: list[str] | None = None
    files: list[MultiFile] | None = None



from __future__ import annotations  # for python 3.8

import os
from typing import BinaryIO, Final

from agstoolbox.core.ags.ags_export import write_uint8, write_bytes, write_uint, write_uint64, \
    write_string_terminated
from agstoolbox.core.utils.file import get_relative_paths, get_size, get_file, mkdirp
from agstoolbox.core.ags.multifilelib import MultiFileLib
from agstoolbox.core.ags.multifile import MultiFile

CLIB_BEGIN_SIGNATURE: Final[bytes] = b"CLIB\x1a"
CLIB_END_SIGNATURE: Final[bytes] = b"CLIB\x01\x02\x03\x04SIGE"
CLIB_VERSION: Final[int] = 30  # large file support, non-encoded
MAXMULTIFILES: Final[int] = 256  # 1-byte index
MAX_PATH: Final[int] = 260  # corresponds to WinAPI definition


# Write asset library header with the table of contents.
# Currently corresponds to writing main lib file in chain in format version 30.
def write_clib_header(our_lib: MultiFileLib, f: BinaryIO):
    write_uint(0, f)  # reserved options
    write_uint(len(our_lib.data_file_names), f)  # reserved options
    for filename in our_lib.data_file_names:
        write_string_terminated(filename, 'utf-8', f)

    write_uint(len(our_lib.files), f)
    for file_info in our_lib.files:
        write_string_terminated(file_info.name, 'utf-8', f)
        write_uint8(file_info.datafile, f)
        write_uint64(file_info.offset, f)
        write_uint64(file_info.length, f)


def get_multifile_lib(files: list[str], files_base_dir: str, base_file_name: str,
                      make_file_name_assumptions: bool) -> MultiFileLib:
    our_lib: MultiFileLib = MultiFileLib()
    our_lib.files = []
    our_lib.data_file_names = []
    current_data_file = 0
    data_assets: list[str] = get_relative_paths(files, files_base_dir)
    file_count = len(files)

    for i in range(file_count):
        file = files[i]
        asset = data_assets[i]
        file_size = get_size(file)
        if len(asset) >= MAX_PATH:
            print(f"Filename too long: {asset}")
            return MultiFileLib()
        m_file: MultiFile = MultiFile()
        m_file.name = asset
        m_file.filename = file
        m_file.datafile = current_data_file
        m_file.length = file_size
        m_file.offset = 0
        our_lib.files.append(m_file)

    # First, set up ourlib.data_filenames array with all the filenames
    # so that write_clib_header will write the correct amount of data
    for i in range(current_data_file + 1):
        if make_file_name_assumptions:
            our_lib.data_file_names.append(f"{base_file_name}.{'ags' if i == 0 else i:03d}")
        else:
            our_lib.data_file_names.append(get_file(base_file_name))

    return our_lib


def find_file_in_path(file_name: str) -> str | None:
    paths_to_try = ["AudioCache", "Speech"]

    if os.path.exists(file_name):
        return file_name

    for path in paths_to_try:
        potential_path = os.path.join(path, file_name)
        if os.path.exists(potential_path):
            return potential_path

    return None


def make_data_file_from_multifile_lib(our_lib: MultiFileLib, out_dir: str):
    mkdirp(out_dir)

    first_data_file_full_path: str = str()
    for i, data_filename in enumerate(our_lib.data_file_names):
        output_file_name = os.path.join(out_dir, data_filename)

        if i == 0:
            first_data_file_full_path = output_file_name

        with open(output_file_name, 'ab') as wout:
            start_offset = wout.tell()
            write_bytes(CLIB_BEGIN_SIGNATURE, wout)
            write_uint8(CLIB_VERSION, wout)
            write_uint8(i, wout)
            if i == 0:
                main_header_offset = wout.tell()
                write_clib_header(our_lib, wout)

            for file_info in our_lib.files:
                if file_info.datafile == i:
                    file_info.offset = wout.tell() - start_offset
                    filename = find_file_in_path(file_info.filename)
                    if filename is None:
                        try:
                            os.remove(output_file_name)
                        except FileNotFoundError:
                            print('ERROR: Unable to find file' + file_info.filename + '.')
                            print('WARN: Do not remove files during the compilation process.')

                    with open(filename, 'rb') as infile:
                        while chunk := infile.read(1024 * 1024):
                            wout.write(chunk)
                        if infile.tell() < file_info.length:
                            raise IOError(
                                f"Error writing file '{file_info.filename}': possibly disk full")

            if start_offset > 0:
                write_uint64(start_offset, wout)
                write_bytes(CLIB_END_SIGNATURE, wout)

        with open(first_data_file_full_path, 'r+b') as wout:
            wout.seek(main_header_offset, os.SEEK_SET)
            write_clib_header(our_lib, wout)
